# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definitions for Groundedness QA metrics."""
from typing import List
from azureml.metrics.common._metric_base import NonScalarMetric
from azureml.metrics.rai.groundedness._groundedness_base import GroundednessBase


class GroundednessQA(GroundednessBase, NonScalarMetric):
    """Groundedness metric for question answering"""

    def construct_user_prompts(self, context: str, gen_content: str) -> List[str]:
        instruction = \
            "You will be presented with a CONTEXT and an ANSWER about that CONTEXT. " \
            "You need to decide whether the ANSWER is entailed by the CONTEXT by choosing one of" \
            " the following rating:\n" \
            "1. 5: The ANSWER follows logically from the information contained in the CONTEXT.\n" \
            "2. 1: The ANSWER is logically false from the information contained in the CONTEXT.\n" \
            "3. an integer score between 1 and 5 and if such integer score does not exist, " \
            "use 1: It is not possible to determine whether the ANSWER is true or false without " \
            "further information. " \
            "Read the passage of information thoroughly and select the correct " \
            "answer from the three answer labels. " \
            "Read the CONTEXT thoroughly to ensure you know what the CONTEXT entails. " \
            "Note the ANSWER is generated by a computer system, it can contain certain symbols, " \
            "which should not be a negative factor in the evaluation.\n"

        examples = \
            "Independent Examples:\n" \
            "## Example Task #1 Input:\n" \
            """{"CONTEXT": "The Academy Awards, also known as the Oscars are awards for """ \
            "artistic and technical merit for the film industry. " \
            "They are presented annually by the Academy of Motion Picture Arts and Sciences, " \
            "in recognition of excellence in cinematic achievements as assessed by the Academy's " \
            "voting membership. " \
            "The Academy Awards are regarded by many as the most prestigious, significant " \
            "awards in the entertainment " \
            """industry in the United States and worldwide.", """ \
            """"QUESTION": "", "ANSWER": "Oscar is presented every other two years"}\n""" \
            "## Example Task #1 Output:\n" \
            "1\n" \
            "## Example Task #2 Input:\n" \
            """{"CONTEXT": "The Academy Awards, also known as the Oscars are awards for """ \
            "artistic and technical merit for the film industry. " \
            "They are presented annually by the Academy of Motion Picture Arts and Sciences, " \
            "in recognition of excellence in cinematic achievements as assessed by the Academy's " \
            "voting membership. " \
            "The Academy Awards are regarded by many as the most prestigious, significant awards in the " \
            "entertainment " \
            """industry in the United States and worldwide.", """ \
            """"QUESTION": "", "ANSWER": "Oscar is very important awards in the entertainment industry in the """ \
            """United States. And it's also significant worldwide"}\n""" \
            "## Example Task #2 Output:\n" \
            "5\n" \
            "## Example Task #3 Input:\n" \
            """{"CONTEXT": "In Quebec, an allophone is a resident, usually an immigrant, whose mother tongue """\
            """or home language is neither French nor English.", """ \
            """"QUESTION": "", "ANSWER": "In Quebec, an allophone is a resident, usually an""" \
            " immigrant, whose mother " \
            """tongue or home language is not French."}\n""" \
            "## Example Task #3 Output:\n" \
            "5\n" \
            "## Example Task #4 Input:\n" \
            """{"CONTEXT": "Some are reported as not having been wanted at all.", "QUESTION": "", """ \
            """"ANSWER": "All are reported as being completely and fully wanted."}\n""" \
            "## Example Task #4 Output:\n" \
            "1\n"

        reminder = self.get_formatting_reminder()

        actual_task = \
            "## Actual Task Input:\n" \
            '{{"CONTEXT": {}, "QUESTION": "", "ANSWER": {}}}\n'.format(context, gen_content)

        output_prompt = "Actual Task Output:"

        prompt = instruction + examples + actual_task + reminder + output_prompt
        return [prompt]
